/*
Copyright 2022 The Matrix.org Foundation C.I.C.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

import { MatrixClient } from "matrix-js-sdk/src/matrix";

import { SdkContextClass } from "../../src/contexts/SDKContext";
import { OidcClientStore } from "../../src/stores/oidc/OidcClientStore";
import { UserProfilesStore } from "../../src/stores/UserProfilesStore";
import { VoiceBroadcastPreRecordingStore } from "../../src/voice-broadcast";
import { createTestClient } from "../test-utils";

jest.mock("../../src/voice-broadcast/stores/VoiceBroadcastPreRecordingStore");

describe("SdkContextClass", () => {
    let sdkContext = SdkContextClass.instance;
    let client: MatrixClient;

    beforeAll(() => {
        client = createTestClient();
    });

    beforeEach(() => {
        sdkContext = new SdkContextClass();
    });

    it("instance should always return the same instance", () => {
        const globalInstance = SdkContextClass.instance;
        expect(SdkContextClass.instance).toBe(globalInstance);
    });

    it("voiceBroadcastPreRecordingStore should always return the same VoiceBroadcastPreRecordingStore", () => {
        const first = sdkContext.voiceBroadcastPreRecordingStore;
        expect(first).toBeInstanceOf(VoiceBroadcastPreRecordingStore);
        expect(sdkContext.voiceBroadcastPreRecordingStore).toBe(first);
    });

    it("userProfilesStore should raise an error without a client", () => {
        expect(() => {
            sdkContext.userProfilesStore;
        }).toThrow("Unable to create UserProfilesStore without a client");
    });

    it("oidcClientStore should raise an error without a client", () => {
        expect(() => {
            sdkContext.oidcClientStore;
        }).toThrow("Unable to create OidcClientStore without a client");
    });

    describe("when SDKContext has a client", () => {
        beforeEach(() => {
            sdkContext.client = client;
        });

        it("userProfilesStore should return a UserProfilesStore", () => {
            const store = sdkContext.userProfilesStore;
            expect(store).toBeInstanceOf(UserProfilesStore);
            // it should return the same instance
            expect(sdkContext.userProfilesStore).toBe(store);
        });

        it("onLoggedOut should clear the UserProfilesStore", () => {
            const store = sdkContext.userProfilesStore;
            sdkContext.onLoggedOut();
            expect(sdkContext.userProfilesStore).not.toBe(store);
        });

        it("oidcClientstore should return a OidcClientStore", () => {
            const store = sdkContext.oidcClientStore;
            expect(store).toBeInstanceOf(OidcClientStore);
            // it should return the same instance
            expect(sdkContext.oidcClientStore).toBe(store);
        });
    });
});
